/*
Module : ShelExec.cpp
Purpose: Implementation of a simple ShellExecute wrapper app

Copyright (c) 1999 - 2016 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 
*/


////////////////////////// Includes ///////////////////////////////////////////

#include "stdafx.h"
#include "ShelExec.h"


////////////////////////// Macros / Defines ///////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

#ifndef SEE_MASK_NOASYNC
#define SEE_MASK_NOASYNC 0x00000100
#endif //#ifndef SEE_MASK_NOASYNC


////////////////////////// Implementation /////////////////////////////////////

class CEnumDisplayArg
{
public:
//Constructors / Destructors
  CEnumDisplayArg() : m_hMon(NULL)
  {
  }

//Member variables
  CString  m_sMonitor; //The name of the monitor to match
  HMONITOR m_hMon;     //If found the monitor handle
};


class CMyCommandLineInfo : public CCommandLineInfo
{
public:
//Constructors / Destructors
  CMyCommandLineInfo();

//Methods
  virtual void ParseParam(const TCHAR* pszParam, BOOL bFlag, BOOL bLast);

//Data
  CString m_sVerb;
  CString m_sDir;
  int     m_nCmdShow;
  BOOL    m_bExe;
  CString m_sParams;
  BOOL    m_bWait;
  BOOL    m_bNoShortFileNameConversion;
  BOOL    m_bNoUI;
  BOOL    m_bWaitForInputIdle;
  CString m_sMonitor;
};


//Callback function used to find the correct monitor handle
BOOL CALLBACK EnumDispProc(HMONITOR hMon, HDC /*dcMon*/, RECT* /*pRcMon*/, LPARAM lParam)
{
  CEnumDisplayArg* pArg = reinterpret_cast<CEnumDisplayArg*>(lParam);
  AFXASSUME(pArg != NULL);

  MONITORINFOEX miex;
  miex.cbSize = sizeof(miex);
  if (GetMonitorInfo(hMon, &miex))
  {
    if (miex.szDevice == pArg->m_sMonitor)
    {
      pArg->m_hMon = hMon;
      return FALSE;
    }
  }

  return TRUE;
}


CMyCommandLineInfo::CMyCommandLineInfo() : m_nCmdShow(SW_SHOW),
                                           m_bExe(FALSE),
                                           m_bWait(FALSE),
                                           m_bNoShortFileNameConversion(FALSE),
                                           m_bNoUI(FALSE),
                                           m_bWaitForInputIdle(FALSE)
{
}

void CMyCommandLineInfo::ParseParam(const TCHAR* pszParam, BOOL bFlag, BOOL bLast)
{
  if (bFlag)
  {
    CString sParam(pszParam);
    CString sUpCaseParam(pszParam);
    sUpCaseParam.MakeUpper();

    //Handle the various command line parameters we support
    int nLength = sUpCaseParam.GetLength();
    if ((sUpCaseParam.Find(_T("SHOWCMD")) == 0) && (nLength > 8))
    {
      CString sShowVal(sParam.Right(nLength - 8));
      m_nCmdShow = _ttoi(sShowVal);
    }
    else if ((sUpCaseParam.Find(_T("VERB")) == 0) && (nLength > 5))
    {
      m_sVerb = sParam.Right(nLength - 5);
    }
    else if ((sUpCaseParam.Find(_T("DIR")) == 0) && (nLength > 4))
    {
      m_sDir = sParam.Right(nLength - 4);
    }
    else if ((sUpCaseParam.Find(_T("PARAMS")) == 0) && (nLength > 7))
    {
      m_sParams = sParam.Right(nLength - 7);
    }
    else if (sUpCaseParam == _T("WAIT"))
    {
      m_bWait = TRUE;
    }
    else if (sUpCaseParam == _T("WAITFORINPUTIDLE"))
    {
      m_bWaitForInputIdle = TRUE;
    }
    else if (sUpCaseParam == _T("NOSFNC"))
    {
      m_bNoShortFileNameConversion = TRUE;
    }
    else if (sParam == _T("EXE"))
    {
      m_bExe = TRUE;
    }
    else if (sUpCaseParam == _T("NOUI"))
    {
      m_bNoUI = TRUE;
    }
    else if ((sUpCaseParam.Find(_T("MONITOR")) == 0) && (nLength > 8))
    {
      m_sMonitor = sParam.Right(nLength - 8);
    }
    else
    {
      //Let the base class do its thing
      __super::ParseParam(pszParam, bFlag, bLast);
    }
  }
  else
  {
    //Let the base class do its thing
    CCommandLineInfo::ParseParam(pszParam, bFlag, bLast);
  }
}


BEGIN_MESSAGE_MAP(CShelExecApp, CWinApp)
END_MESSAGE_MAP()

CShelExecApp::CShelExecApp() : m_nReturnCode(ERROR_SUCCESS)
{
}

CShelExecApp theApp;


BOOL CShelExecApp::InitInstance()
{
  //Parse the command line
  CMyCommandLineInfo cmdInfo;
  ParseCommandLine(cmdInfo);

  //Show the usage if no filename was specified
  if (cmdInfo.m_strFileName.IsEmpty())
  {
    m_nReturnCode = ERROR_INVALID_PARAMETER;

  #ifdef _CONSOLE
    if (!cmdInfo.m_bNoUI)
    {
      CString sMsg;
      BOOL bLoaded = sMsg.LoadString(IDP_USAGE);
      UNREFERENCED_PARAMETER(bLoaded);
      _tprintf(_T("%s\n"), sMsg.operator LPCTSTR());
    }
  #else
    if (!cmdInfo.m_bNoUI)
      AfxMessageBox(IDP_USAGE);
  #endif //#ifdef _CONSOLE
    return FALSE;
  }

  //Find the current diretory and drive for this instance
  TCHAR szPath[_MAX_PATH];
  szPath[0] = _T('\0');
  if (!GetModuleFileName(NULL, szPath, _MAX_PATH))
  {
    m_nReturnCode = GetLastError();

    //Report the error if required
    CString sErrorCode;
    sErrorCode.Format(_T("%u"), m_nReturnCode);
    CString sMsg;
    AfxFormatString2(sMsg, IDP_FAIL_GETMODULEFILENAME, sErrorCode, FormatMessage(m_nReturnCode));
  #ifdef _CONSOLE
    if (!cmdInfo.m_bNoUI)
      _tprintf(_T("%s\n"), sMsg.operator LPCTSTR());
  #else   
    if (!cmdInfo.m_bNoUI)
      AfxMessageBox(sMsg, MB_OK | MB_ICONEXCLAMATION);
  #endif //#ifdef _CONSOLE
    return FALSE;
  }
  TCHAR szDrive[_MAX_DRIVE];
  szDrive[0] = _T('\0');
  TCHAR szDir[_MAX_DIR];
  szDir[0] = _T('\0');
  TCHAR szFolder[_MAX_PATH];
  szFolder[0] = _T('\0');
  _tsplitpath_s(szPath, szDrive, sizeof(szDrive)/sizeof(TCHAR), szDir, sizeof(szDir)/sizeof(TCHAR), NULL, 0, NULL, 0);
  _tmakepath_s(szFolder, sizeof(szFolder)/sizeof(TCHAR), szDrive, szDir, NULL, NULL);
  CString sFolder(szFolder);
  int nLength = sFolder.GetLength();
  if (nLength && sFolder.GetAt(nLength-1) == _T('\\'))
    sFolder = sFolder.Left(nLength-1);
  CString sDrive(szDrive);
  nLength = sDrive.GetLength();
  if (nLength && sDrive.GetAt(nLength-1) == _T('\\'))
    sDrive = sDrive.Left(nLength-1);

  //Replace any variables with their actual values as determined above
  cmdInfo.m_sDir.Replace(_T("@EXEDRV@"), sDrive);
  cmdInfo.m_sParams.Replace(_T("@EXEDRV@"), sDrive);
  cmdInfo.m_strFileName.Replace(_T("@EXEDRV@"), sDrive);

  cmdInfo.m_sDir.Replace(_T("@EXEDIR@"), sFolder);
  cmdInfo.m_sParams.Replace(_T("@EXEDIR@"), sFolder);
  cmdInfo.m_strFileName.Replace(_T("@EXEDIR@"), sFolder);

  cmdInfo.m_sDir.Replace(_T("@QUOTE@"), _T("\""));
  cmdInfo.m_sParams.Replace(_T("@QUOTE@"), _T("\""));
  cmdInfo.m_strFileName.Replace(_T("@QUOTE@"), _T("\""));


  //The structure which will be passed to ShellExecuteEx
  SHELLEXECUTEINFO sei;
  memset(&sei, 0, sizeof(sei));
  sei.cbSize = sizeof(sei);

  //Convert the filename to its short version if we did not specify no short filename conversion
  if (!cmdInfo.m_bNoShortFileNameConversion)
  {
    //convert to short filename format if we can
    if (GetShortPathName(cmdInfo.m_strFileName, szPath, _MAX_PATH) != 0)
      cmdInfo.m_strFileName = szPath;
  }

  //Setup the structure with common parameters
  sei.nShow = cmdInfo.m_nCmdShow;
  sei.lpFile = cmdInfo.m_strFileName.GetBuffer(cmdInfo.m_strFileName.GetLength());
  sei.fMask = SEE_MASK_DOENVSUBST | SEE_MASK_NOASYNC | SEE_MASK_INVOKEIDLIST;
  if (cmdInfo.m_bWait)
    sei.fMask |= SEE_MASK_NOCLOSEPROCESS;
  if (cmdInfo.m_bNoUI)
    sei.fMask |= SEE_MASK_FLAG_NO_UI;
  if (cmdInfo.m_bWaitForInputIdle)
    sei.fMask |= SEE_MASK_WAITFORINPUTIDLE;

  //Get the monitor handle if required
  if (cmdInfo.m_sMonitor.GetLength())
  {
    CEnumDisplayArg dispArg;
    dispArg.m_sMonitor = cmdInfo.m_sMonitor;
    EnumDisplayMonitors(NULL, NULL, EnumDispProc, reinterpret_cast<LPARAM>(&dispArg));
    if (dispArg.m_hMon == NULL)
    {
      m_nReturnCode = ERROR_FILE_NOT_FOUND;

      //Report the error if required
      CString sMsg;
      AfxFormatString1(sMsg, IDP_FAIL_FIND_MONITOR, cmdInfo.m_sMonitor);
    #ifdef _CONSOLE
      if (!cmdInfo.m_bNoUI)
        _tprintf(_T("%s\n"), sMsg.operator LPCTSTR());
    #else
      if (!cmdInfo.m_bNoUI)
        AfxMessageBox(sMsg, MB_OK | MB_ICONEXCLAMATION);
    #endif //#ifdef _CONSOLE
      return FALSE;
    }

    sei.hMonitor = dispArg.m_hMon;
    sei.fMask |= SEE_MASK_HMONITOR;
  }

  //Did we specify a working directory
  if (cmdInfo.m_sDir.GetLength())
    sei.lpDirectory = cmdInfo.m_sDir.GetBuffer(cmdInfo.m_sDir.GetLength());

  //Set any optional parameters
  if (cmdInfo.m_bExe)
  {
    if (cmdInfo.m_sParams.GetLength())
      sei.lpParameters = cmdInfo.m_sParams.GetBuffer(cmdInfo.m_sParams.GetLength());
  }
  else
  {
    if (cmdInfo.m_sVerb.GetLength())
      sei.lpVerb = cmdInfo.m_sVerb.GetBuffer(cmdInfo.m_sVerb.GetLength());
  }


  //Call ShellExecuteEx API
  ShellExecuteEx(&sei);
  m_nReturnCode = GetLastError();

  //Release all our CString buffers
  cmdInfo.m_strFileName.ReleaseBuffer();
  if (cmdInfo.m_sDir.GetLength())
    cmdInfo.m_sDir.ReleaseBuffer();
  if (cmdInfo.m_bExe)
  {
    if (cmdInfo.m_sParams.GetLength())
      cmdInfo.m_sParams.ReleaseBuffer();
  }
  else
  {
    if (cmdInfo.m_sVerb.GetLength())
      cmdInfo.m_sVerb.ReleaseBuffer();
  }

  //Wait for the process to exit if required
  if (cmdInfo.m_bWait)
  {
    if (sei.hProcess != 0)
      WaitForSingleObject(sei.hProcess, INFINITE);
  }

  //Close the process handle if required
  if (sei.hProcess != 0)
    CloseHandle(sei.hProcess);

  return FALSE;

}

int CShelExecApp::ExitInstance()
{
  //Let the base class do its thing
  __super::ExitInstance();

  //Return the error code we have cached
  return m_nReturnCode;
}

CString CShelExecApp::FormatMessage(DWORD dwError)
{
  //What will be the return value from this method
  CString sError;

  //Lookup the error using FormatMessage
  LPTSTR lpszError = NULL;
  DWORD dwReturn = ::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                                   NULL, dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT), reinterpret_cast<LPTSTR>(&lpszError), 0, NULL);
  if (dwReturn)
  {
    sError = lpszError;
    LocalFree(lpszError);
  }
  
  return sError;
}